#include "moar.h"
/* This file is generated from src/core/oplist by tools/update_ops.raku. */
static const MVMOpInfo MVM_op_infos[] = {
    {
        MVM_OP_no_op,
        "no_op",
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_const_i8,
        "const_i8",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int8, MVM_operand_int8 }
    },
    {
        MVM_OP_const_i16,
        "const_i16",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int16, MVM_operand_int16 }
    },
    {
        MVM_OP_const_i32,
        "const_i32",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int32, MVM_operand_int32 }
    },
    {
        MVM_OP_const_i64,
        "const_i64",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_int64 }
    },
    {
        MVM_OP_const_n32,
        "const_n32",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num32, MVM_operand_num32 }
    },
    {
        MVM_OP_const_n64,
        "const_n64",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_num64 }
    },
    {
        MVM_OP_const_s,
        "const_s",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_str }
    },
    {
        MVM_OP_set,
        "set",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_type_var, MVM_operand_read_reg | MVM_operand_type_var }
    },
    {
        MVM_OP_extend_u8,
        "extend_u8",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint8 }
    },
    {
        MVM_OP_extend_u16,
        "extend_u16",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint16 }
    },
    {
        MVM_OP_extend_u32,
        "extend_u32",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint32 }
    },
    {
        MVM_OP_extend_i8,
        "extend_i8",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int8 }
    },
    {
        MVM_OP_extend_i16,
        "extend_i16",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int16 }
    },
    {
        MVM_OP_extend_i32,
        "extend_i32",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int32 }
    },
    {
        MVM_OP_trunc_u8,
        "trunc_u8",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint8, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_trunc_u16,
        "trunc_u16",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint16, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_trunc_u32,
        "trunc_u32",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint32, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_trunc_i8,
        "trunc_i8",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int8, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_trunc_i16,
        "trunc_i16",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_trunc_i32,
        "trunc_i32",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int32, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_extend_n32,
        "extend_n32",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num32 }
    },
    {
        MVM_OP_trunc_n32,
        "trunc_n32",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num32, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_goto,
        "goto",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_ins }
    },
    {
        MVM_OP_if_i,
        "if_i",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_int64, MVM_operand_ins }
    },
    {
        MVM_OP_unless_i,
        "unless_i",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_int64, MVM_operand_ins }
    },
    {
        MVM_OP_if_n,
        "if_n",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_num64, MVM_operand_ins }
    },
    {
        MVM_OP_unless_n,
        "unless_n",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_num64, MVM_operand_ins }
    },
    {
        MVM_OP_if_s,
        "if_s",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_unless_s,
        "unless_s",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_if_s0,
        "if_s0",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_unless_s0,
        "unless_s0",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_eq_u,
        "eq_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_ne_u,
        "ne_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_jumplist,
        "jumplist",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_getlex,
        "getlex",
        2,
        1,
        1,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_type_var, MVM_operand_read_lex | MVM_operand_type_var }
    },
    {
        MVM_OP_bindlex,
        "bindlex",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_lex | MVM_operand_type_var, MVM_operand_read_reg | MVM_operand_type_var }
    },
    {
        MVM_OP_getlex_ni,
        "getlex_ni",
        2,
        1,
        2,
        0,
        0,
        0,
        1,
        3,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_str }
    },
    {
        MVM_OP_getlex_nn,
        "getlex_nn",
        2,
        1,
        2,
        0,
        0,
        0,
        1,
        3,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_str }
    },
    {
        MVM_OP_getlex_ns,
        "getlex_ns",
        2,
        1,
        2,
        0,
        0,
        0,
        1,
        3,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_str }
    },
    {
        MVM_OP_getlex_no,
        "getlex_no",
        2,
        1,
        3,
        0,
        1,
        0,
        1,
        3,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str }
    },
    {
        MVM_OP_bindlex_ni,
        "bindlex_ni",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bindlex_nn,
        "bindlex_nn",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_str, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_bindlex_ns,
        "bindlex_ns",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindlex_no,
        "bindlex_no",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getlex_ng,
        "getlex_ng",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindlex_ng,
        "bindlex_ng",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getdynlex,
        "getdynlex",
        2,
        1,
        2,
        0,
        0,
        0,
        1,
        3,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_binddynlex,
        "binddynlex",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_setlexvalue,
        "setlexvalue",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_lexprimspec,
        "lexprimspec",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_return_i,
        "return_i",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_return_n,
        "return_n",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_return_s,
        "return_s",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_return_o,
        "return_o",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_return,
        "return",
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_eq_i,
        "eq_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_ne_i,
        "ne_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_lt_i,
        "lt_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_le_i,
        "le_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_gt_i,
        "gt_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_ge_i,
        "ge_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_cmp_i,
        "cmp_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_add_i,
        "add_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_sub_i,
        "sub_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_mul_i,
        "mul_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_div_i,
        "div_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_div_u,
        "div_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_mod_i,
        "mod_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_mod_u,
        "mod_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_neg_i,
        "neg_i",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_abs_i,
        "abs_i",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_inc_i,
        "inc_i",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_inc_u,
        "inc_u",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_dec_i,
        "dec_i",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_dec_u,
        "dec_u",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_band_i,
        "band_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bor_i,
        "bor_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bxor_i,
        "bxor_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bnot_i,
        "bnot_i",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_blshift_i,
        "blshift_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_brshift_i,
        "brshift_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_pow_i,
        "pow_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_not_i,
        "not_i",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_gcd_i,
        "gcd_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_lcm_i,
        "lcm_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_eq_n,
        "eq_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_ne_n,
        "ne_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_lt_n,
        "lt_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_le_n,
        "le_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_gt_n,
        "gt_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_ge_n,
        "ge_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_cmp_n,
        "cmp_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_add_n,
        "add_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_sub_n,
        "sub_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_mul_n,
        "mul_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_div_n,
        "div_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_mod_n,
        "mod_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_neg_n,
        "neg_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_abs_n,
        "abs_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_pow_n,
        "pow_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_ceil_n,
        "ceil_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_floor_n,
        "floor_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_sin_n,
        "sin_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_asin_n,
        "asin_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_cos_n,
        "cos_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_acos_n,
        "acos_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_tan_n,
        "tan_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_atan_n,
        "atan_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_atan2_n,
        "atan2_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_le_u,
        "le_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_lt_u,
        "lt_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_sinh_n,
        "sinh_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_cosh_n,
        "cosh_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_tanh_n,
        "tanh_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_ge_u,
        "ge_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_sqrt_n,
        "sqrt_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_log_n,
        "log_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_exp_n,
        "exp_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_coerce_in,
        "coerce_in",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_coerce_ni,
        "coerce_ni",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_coerce_is,
        "coerce_is",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_coerce_ns,
        "coerce_ns",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_coerce_si,
        "coerce_si",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_coerce_sn,
        "coerce_sn",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_gt_u,
        "gt_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_cmp_u,
        "cmp_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_DEPRECATED_99,
        "DEPRECATED_99",
        1,
        0,
        8,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_callsite }
    },
    {
        MVM_OP_DEPRECATED_100,
        "DEPRECATED_100",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_DEPRECATED_101,
        "DEPRECATED_101",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_int16, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_DEPRECATED_102,
        "DEPRECATED_102",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_int16, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_DEPRECATED_103,
        "DEPRECATED_103",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_int16, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_104,
        "DEPRECATED_104",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_int16, MVM_operand_int64 }
    },
    {
        MVM_OP_DEPRECATED_105,
        "DEPRECATED_105",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_int16, MVM_operand_num64 }
    },
    {
        MVM_OP_DEPRECATED_106,
        "DEPRECATED_106",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_int16, MVM_operand_str }
    },
    {
        MVM_OP_DEPRECATED_70,
        "DEPRECATED_70",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_71,
        "DEPRECATED_71",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_72,
        "DEPRECATED_72",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_73,
        "DEPRECATED_73",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_74,
        "DEPRECATED_74",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_checkarity,
        "checkarity",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_int16, MVM_operand_int16 }
    },
    {
        MVM_OP_param_rp_i,
        "param_rp_i",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_int16 }
    },
    {
        MVM_OP_param_rp_n,
        "param_rp_n",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_int16 }
    },
    {
        MVM_OP_param_rp_s,
        "param_rp_s",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_param_rp_o,
        "param_rp_o",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_param_op_i,
        "param_op_i",
        3,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_int16, MVM_operand_ins }
    },
    {
        MVM_OP_param_op_n,
        "param_op_n",
        3,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_int16, MVM_operand_ins }
    },
    {
        MVM_OP_param_op_s,
        "param_op_s",
        3,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_int16, MVM_operand_ins }
    },
    {
        MVM_OP_param_op_o,
        "param_op_o",
        3,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_ins }
    },
    {
        MVM_OP_param_rn_i,
        "param_rn_i",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_str }
    },
    {
        MVM_OP_param_rn_n,
        "param_rn_n",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_str }
    },
    {
        MVM_OP_param_rn_s,
        "param_rn_s",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_str }
    },
    {
        MVM_OP_param_rn_o,
        "param_rn_o",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str }
    },
    {
        MVM_OP_param_on_i,
        "param_on_i",
        3,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_param_on_n,
        "param_on_n",
        3,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_param_on_s,
        "param_on_s",
        3,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_param_on_o,
        "param_on_o",
        3,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_param_sp,
        "param_sp",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_param_sn,
        "param_sn",
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getcode,
        "getcode",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_coderef }
    },
    {
        MVM_OP_caller,
        "caller",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_capturelex,
        "capturelex",
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_takeclosure,
        "takeclosure",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_exception,
        "exception",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bindexmessage,
        "bindexmessage",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindexpayload,
        "bindexpayload",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bindexcategory,
        "bindexcategory",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_getexmessage,
        "getexmessage",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getexpayload,
        "getexpayload",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getexcategory,
        "getexcategory",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_throwdyn,
        "throwdyn",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_throwlex,
        "throwlex",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_throwlexotic,
        "throwlexotic",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_throwcatdyn,
        "throwcatdyn",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int64 }
    },
    {
        MVM_OP_throwcatlex,
        "throwcatlex",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int64 }
    },
    {
        MVM_OP_throwcatlexotic,
        "throwcatlexotic",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int64 }
    },
    {
        MVM_OP_die,
        "die",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_rethrow,
        "rethrow",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_resume,
        "resume",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_takehandlerresult,
        "takehandlerresult",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_33,
        "DEPRECATED_33",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_ins }
    },
    {
        MVM_OP_DEPRECATED_34,
        "DEPRECATED_34",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_backtracestrings,
        "backtracestrings",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_usecapture,
        "usecapture",
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_savecapture,
        "savecapture",
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_captureposelems,
        "captureposelems",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_captureposarg,
        "captureposarg",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_captureposarg_i,
        "captureposarg_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_captureposarg_n,
        "captureposarg_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_captureposarg_s,
        "captureposarg_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_captureposprimspec,
        "captureposprimspec",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_captureexistsnamed,
        "captureexistsnamed",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_capturehasnameds,
        "capturehasnameds",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_69,
        "DEPRECATED_69",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_60,
        "DEPRECATED_60",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_61,
        "DEPRECATED_61",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_null_s,
        "null_s",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str }
    },
    {
        MVM_OP_isnull_s,
        "isnull_s",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_eq_s,
        "eq_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_ne_s,
        "ne_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_gt_s,
        "gt_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_ge_s,
        "ge_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_lt_s,
        "lt_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_le_s,
        "le_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_cmp_s,
        "cmp_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_eqat_s,
        "eqat_s",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_eqatic_s,
        "eqatic_s",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_haveat_s,
        "haveat_s",
        6,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_concat_s,
        "concat_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_repeat_s,
        "repeat_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_substr_s,
        "substr_s",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_index_s,
        "index_s",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_DEPRECATED_40,
        "DEPRECATED_40",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_codes_s,
        "codes_s",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getcp_s,
        "getcp_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_indexcp_s,
        "indexcp_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_uc,
        "uc",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_lc,
        "lc",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_tc,
        "tc",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_split,
        "split",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_join,
        "join",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getcpbyname,
        "getcpbyname",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_indexat,
        "indexat",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_indexnat,
        "indexnat",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_unipropcode,
        "unipropcode",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_unipvalcode,
        "unipvalcode",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_hasuniprop,
        "hasuniprop",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_hasunipropc,
        "hasunipropc",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_int16, MVM_operand_int16 }
    },
    {
        MVM_OP_chars,
        "chars",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_chr,
        "chr",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_ordfirst,
        "ordfirst",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_ordat,
        "ordat",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_rindexfrom,
        "rindexfrom",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_escape,
        "escape",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_flip,
        "flip",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_setbuffersize_fh,
        "setbuffersize_fh",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_iscclass,
        "iscclass",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_findcclass,
        "findcclass",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_findnotcclass,
        "findnotcclass",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_nfafromstatelist,
        "nfafromstatelist",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_nfarunproto,
        "nfarunproto",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_nfarunalt,
        "nfarunalt",
        6,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_radix,
        "radix",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_encode,
        "encode",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decode,
        "decode",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_istrue_s,
        "istrue_s",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_isfalse_s,
        "isfalse_s",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_null,
        "null",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_isnull,
        "isnull",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_ifnonnull,
        "ifnonnull",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_ins }
    },
    {
        MVM_OP_DEPRECATED_75,
        "DEPRECATED_75",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str }
    },
    {
        MVM_OP_DEPRECATED_76,
        "DEPRECATED_76",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_DEPRECATED_77,
        "DEPRECATED_77",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str }
    },
    {
        MVM_OP_DEPRECATED_78,
        "DEPRECATED_78",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_create,
        "create",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_clone,
        "clone",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_isconcrete,
        "isconcrete",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_rebless,
        "rebless",
        3,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_istype,
        "istype",
        3,
        1,
        10,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_objprimspec,
        "objprimspec",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_gethow,
        "gethow",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getwhat,
        "getwhat",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getwho,
        "getwho",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_setwho,
        "setwho",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_reprname,
        "reprname",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getwhere,
        "getwhere",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_eqaddr,
        "eqaddr",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bindattr_i,
        "bindattr_i",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_int16 }
    },
    {
        MVM_OP_bindattr_n,
        "bindattr_n",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_int16 }
    },
    {
        MVM_OP_bindattr_s,
        "bindattr_s",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_bindattr_o,
        "bindattr_o",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_bindattrs_i,
        "bindattrs_i",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bindattrs_n,
        "bindattrs_n",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_bindattrs_s,
        "bindattrs_s",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindattrs_o,
        "bindattrs_o",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getattr_i,
        "getattr_i",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_getattr_n,
        "getattr_n",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_getattr_s,
        "getattr_s",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_getattr_o,
        "getattr_o",
        5,
        0,
        1,
        0,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_getattrs_i,
        "getattrs_i",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getattrs_n,
        "getattrs_n",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getattrs_s,
        "getattrs_s",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getattrs_o,
        "getattrs_o",
        4,
        0,
        1,
        0,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_attrinited,
        "attrinited",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_box_i,
        "box_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_box_n,
        "box_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_box_s,
        "box_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_unbox_i,
        "unbox_i",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_unbox_n,
        "unbox_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_unbox_s,
        "unbox_s",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_atpos_i,
        "atpos_i",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atpos_n,
        "atpos_n",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atpos_s,
        "atpos_s",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atpos_o,
        "atpos_o",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bindpos_i,
        "bindpos_i",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bindpos_n,
        "bindpos_n",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_bindpos_s,
        "bindpos_s",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindpos_o,
        "bindpos_o",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_push_i,
        "push_i",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_push_n,
        "push_n",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_push_s,
        "push_s",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_push_o,
        "push_o",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_pop_i,
        "pop_i",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_pop_n,
        "pop_n",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_pop_s,
        "pop_s",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_pop_o,
        "pop_o",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_shift_i,
        "shift_i",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_shift_n,
        "shift_n",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_shift_s,
        "shift_s",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_shift_o,
        "shift_o",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_unshift_i,
        "unshift_i",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_unshift_n,
        "unshift_n",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_unshift_s,
        "unshift_s",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_unshift_o,
        "unshift_o",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_splice,
        "splice",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_setelemspos,
        "setelemspos",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_existspos,
        "existspos",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atkey_i,
        "atkey_i",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_atkey_n,
        "atkey_n",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_atkey_s,
        "atkey_s",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_atkey_o,
        "atkey_o",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindkey_i,
        "bindkey_i",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bindkey_n,
        "bindkey_n",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_bindkey_s,
        "bindkey_s",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindkey_o,
        "bindkey_o",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_existskey,
        "existskey",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_deletekey,
        "deletekey",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_elems,
        "elems",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_knowhow,
        "knowhow",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_knowhowattr,
        "knowhowattr",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_newtype,
        "newtype",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_composetype,
        "composetype",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_83,
        "DEPRECATED_83",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_84,
        "DEPRECATED_84",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_settypecache,
        "settypecache",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_settypecheckmode,
        "settypecheckmode",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_setboolspec,
        "setboolspec",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_64,
        "DEPRECATED_64",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_65,
        "DEPRECATED_65",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bootint,
        "bootint",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bootnum,
        "bootnum",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bootstr,
        "bootstr",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bootarray,
        "bootarray",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bootintarray,
        "bootintarray",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bootnumarray,
        "bootnumarray",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bootstrarray,
        "bootstrarray",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_boothash,
        "boothash",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_isint,
        "isint",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_isnum,
        "isnum",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_isstr,
        "isstr",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_islist,
        "islist",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_ishash,
        "ishash",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sethllconfig,
        "sethllconfig",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_hllboxtype_i,
        "hllboxtype_i",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_hllboxtype_n,
        "hllboxtype_n",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_hllboxtype_s,
        "hllboxtype_s",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_hlllist,
        "hlllist",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_hllhash,
        "hllhash",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getcomp,
        "getcomp",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindcomp,
        "bindcomp",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getcurhllsym,
        "getcurhllsym",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindcurhllsym,
        "bindcurhllsym",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_gethllsym,
        "gethllsym",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindhllsym,
        "bindhllsym",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_settypehll,
        "settypehll",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_settypehllrole,
        "settypehllrole",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_DEPRECATED_95,
        "DEPRECATED_95",
        2,
        1,
        0,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_96,
        "DEPRECATED_96",
        3,
        1,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_usecompileehllconfig,
        "usecompileehllconfig",
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_usecompilerhllconfig,
        "usecompilerhllconfig",
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_iter,
        "iter",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_iterkey_s,
        "iterkey_s",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_iterval,
        "iterval",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getcodename,
        "getcodename",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_iscoderef,
        "iscoderef",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getcodeobj,
        "getcodeobj",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_setcodeobj,
        "setcodeobj",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_setcodename,
        "setcodename",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_forceouterctx,
        "forceouterctx",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_97,
        "DEPRECATED_97",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_98,
        "DEPRECATED_98",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_freshcoderef,
        "freshcoderef",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_markcodestatic,
        "markcodestatic",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_markcodestub,
        "markcodestub",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getstaticcode,
        "getstaticcode",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getcodecuid,
        "getcodecuid",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_90,
        "DEPRECATED_90",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_91,
        "DEPRECATED_91",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_assign,
        "assign",
        2,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_assignunchecked,
        "assignunchecked",
        2,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_iscont,
        "iscont",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decont,
        "decont",
        2,
        1,
        3,
        1,
        1,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_setcontspec,
        "setcontspec",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sha1,
        "sha1",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_createsc,
        "createsc",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_scsetobj,
        "scsetobj",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_scsetcode,
        "scsetcode",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_scgetobj,
        "scgetobj",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_scgethandle,
        "scgethandle",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_scgetobjidx,
        "scgetobjidx",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_scsetdesc,
        "scsetdesc",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_scobjcount,
        "scobjcount",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_setobjsc,
        "setobjsc",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getobjsc,
        "getobjsc",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_serialize,
        "serialize",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_deserialize,
        "deserialize",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_wval,
        "wval",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_int16 }
    },
    {
        MVM_OP_wval_wide,
        "wval_wide",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_int64 }
    },
    {
        MVM_OP_scwbdisable,
        "scwbdisable",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_scwbenable,
        "scwbenable",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_pushcompsc,
        "pushcompsc",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_popcompsc,
        "popcompsc",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_scgetdesc,
        "scgetdesc",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_loadbytecode,
        "loadbytecode",
        2,
        0,
        3,
        1,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_newmixintype,
        "newmixintype",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_installconfprog,
        "installconfprog",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_iscompunit,
        "iscompunit",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_compunitmainline,
        "compunitmainline",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_compunitcodes,
        "compunitcodes",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_ctx,
        "ctx",
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_ctxouter,
        "ctxouter",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_ctxcaller,
        "ctxcaller",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_ctxlexpad,
        "ctxlexpad",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_curcode,
        "curcode",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_callercode,
        "callercode",
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_add_I,
        "add_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sub_I,
        "sub_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_mul_I,
        "mul_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_div_I,
        "div_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_mod_I,
        "mod_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_neg_I,
        "neg_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_abs_I,
        "abs_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_cmp_I,
        "cmp_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_eq_I,
        "eq_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_ne_I,
        "ne_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_lt_I,
        "lt_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_le_I,
        "le_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_gt_I,
        "gt_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_ge_I,
        "ge_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bor_I,
        "bor_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bxor_I,
        "bxor_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_band_I,
        "band_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bnot_I,
        "bnot_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_blshift_I,
        "blshift_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_brshift_I,
        "brshift_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_pow_I,
        "pow_I",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_gcd_I,
        "gcd_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_lcm_I,
        "lcm_I",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_expmod_I,
        "expmod_I",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_isprime_I,
        "isprime_I",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_rand_I,
        "rand_I",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_coerce_In,
        "coerce_In",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_coerce_Is,
        "coerce_Is",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_coerce_nI,
        "coerce_nI",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_coerce_sI,
        "coerce_sI",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_isbig_I,
        "isbig_I",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bool_I,
        "bool_I",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_base_I,
        "base_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_radix_I,
        "radix_I",
        6,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_div_In,
        "div_In",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_copy_f,
        "copy_f",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_append_f,
        "append_f",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_rename_f,
        "rename_f",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_delete_f,
        "delete_f",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_chmod_f,
        "chmod_f",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_exists_f,
        "exists_f",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_mkdir,
        "mkdir",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_rmdir,
        "rmdir",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_open_dir,
        "open_dir",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_read_dir,
        "read_dir",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_close_dir,
        "close_dir",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_open_fh,
        "open_fh",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_close_fh,
        "close_fh",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_dispatch_u,
        "dispatch_u",
        3,
        0,
        10,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_str, MVM_operand_callsite }
    },
    {
        MVM_OP_DEPRECATED_27,
        "DEPRECATED_27",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_DEPRECATED_28,
        "DEPRECATED_28",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_atposnd_u,
        "atposnd_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_seek_fh,
        "seek_fh",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_lock_fh,
        "lock_fh",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_unlock_fh,
        "unlock_fh",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sync_fh,
        "sync_fh",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_trunc_fh,
        "trunc_fh",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_eof_fh,
        "eof_fh",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getstdin,
        "getstdin",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getstdout,
        "getstdout",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getstderr,
        "getstderr",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_connect_sk,
        "connect_sk",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint16 }
    },
    {
        MVM_OP_socket,
        "socket",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bind_sk,
        "bind_sk",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint16, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_coerce_su,
        "coerce_su",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_accept_sk,
        "accept_sk",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decodetocodes,
        "decodetocodes",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_encodefromcodes,
        "encodefromcodes",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_multidimref_u,
        "multidimref_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_print,
        "print",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_say,
        "say",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_return_u,
        "return_u",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_tell_fh,
        "tell_fh",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_stat,
        "stat",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_DEPRECATED_79,
        "DEPRECATED_79",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str }
    },
    {
        MVM_OP_DEPRECATED_80,
        "DEPRECATED_80",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_chdir,
        "chdir",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_srand,
        "srand",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_rand_i,
        "rand_i",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_rand_n,
        "rand_n",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_DEPRECATED_46,
        "DEPRECATED_46",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_sleep,
        "sleep",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_newthread,
        "newthread",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_threadjoin,
        "threadjoin",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_47,
        "DEPRECATED_47",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_exit,
        "exit",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_DEPRECATED_30,
        "DEPRECATED_30",
        8,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_cwd,
        "cwd",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str }
    },
    {
        MVM_OP_clargs,
        "clargs",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getenvhash,
        "getenvhash",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_loadlib,
        "loadlib",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_freelib,
        "freelib",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_findsym,
        "findsym",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_dropsym,
        "dropsym",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_loadext,
        "loadext",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_backendconfig,
        "backendconfig",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getlexouter,
        "getlexouter",
        2,
        1,
        2,
        0,
        0,
        0,
        1,
        3,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getlexrel,
        "getlexrel",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getlexreldyn,
        "getlexreldyn",
        3,
        1,
        2,
        0,
        0,
        0,
        0,
        3,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getlexrelcaller,
        "getlexrelcaller",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getlexcaller,
        "getlexcaller",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bitand_s,
        "bitand_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bitor_s,
        "bitor_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bitxor_s,
        "bitxor_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_isnanorinf,
        "isnanorinf",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_inf,
        "inf",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_neginf,
        "neginf",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_nan,
        "nan",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_getpid,
        "getpid",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_DEPRECATED_29,
        "DEPRECATED_29",
        8,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_filereadable,
        "filereadable",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_filewritable,
        "filewritable",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_fileexecutable,
        "fileexecutable",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_atpos2d_u,
        "atpos2d_u",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_capturenamedshash,
        "capturenamedshash",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_read_fhb,
        "read_fhb",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_write_fhb,
        "write_fhb",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_replace,
        "replace",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_newexception,
        "newexception",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_permit,
        "permit",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_backtrace,
        "backtrace",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_symlink,
        "symlink",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_link,
        "link",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_gethostname,
        "gethostname",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str }
    },
    {
        MVM_OP_exreturnafterunwind,
        "exreturnafterunwind",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_vmeventsubscribe,
        "vmeventsubscribe",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_continuationreset,
        "continuationreset",
        3,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_continuationcontrol,
        "continuationcontrol",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_continuationinvoke,
        "continuationinvoke",
        3,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_randscale_n,
        "randscale_n",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_uniisblock,
        "uniisblock",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_assertparamcheck,
        "assertparamcheck",
        1,
        0,
        2,
        1,
        0,
        0,
        1,
        1,
        0,
        0,
        1,
        { MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_hintfor,
        "hintfor",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_paramnamesused,
        "paramnamesused",
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_getuniname,
        "getuniname",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_getuniprop_int,
        "getuniprop_int",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_getuniprop_bool,
        "getuniprop_bool",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_getuniprop_str,
        "getuniprop_str",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_matchuniprop,
        "matchuniprop",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_nativecallbuild,
        "nativecallbuild",
        7,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_nativecallinvoke,
        "nativecallinvoke",
        4,
        0,
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_nativecallrefresh,
        "nativecallrefresh",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_threadrun,
        "threadrun",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_threadid,
        "threadid",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_threadyield,
        "threadyield",
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_currentthread,
        "currentthread",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_lock,
        "lock",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_unlock,
        "unlock",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_semacquire,
        "semacquire",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_semtryacquire,
        "semtryacquire",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_semrelease,
        "semrelease",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getlockcondvar,
        "getlockcondvar",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_condwait,
        "condwait",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_condsignalone,
        "condsignalone",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_condsignalall,
        "condsignalall",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_queuepoll,
        "queuepoll",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_82,
        "DEPRECATED_82",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_ctxouterskipthunks,
        "ctxouterskipthunks",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_ctxcallerskipthunks,
        "ctxcallerskipthunks",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_timer,
        "timer",
        6,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_cancel,
        "cancel",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_signal,
        "signal",
        7,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_watchfile,
        "watchfile",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_asyncconnect,
        "asyncconnect",
        6,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_asynclisten,
        "asynclisten",
        7,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bindposnd_u,
        "bindposnd_u",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_asyncwritebytes,
        "asyncwritebytes",
        6,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bindpos3d_u,
        "bindpos3d_u",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_asyncreadbytes,
        "asyncreadbytes",
        6,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getlexstatic_o,
        "getlexstatic_o",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        1,
        1,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getlexperinvtype_o,
        "getlexperinvtype_o",
        2,
        1,
        2,
        0,
        1,
        0,
        0,
        3,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_execname,
        "execname",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str }
    },
    {
        MVM_OP_const_i64_16,
        "const_i64_16",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_int16 }
    },
    {
        MVM_OP_const_i64_32,
        "const_i64_32",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_int32 }
    },
    {
        MVM_OP_isnonnull,
        "isnonnull",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_param_rn2_i,
        "param_rn2_i",
        3,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_str, MVM_operand_str }
    },
    {
        MVM_OP_param_rn2_n,
        "param_rn2_n",
        3,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_str, MVM_operand_str }
    },
    {
        MVM_OP_param_rn2_s,
        "param_rn2_s",
        3,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_str, MVM_operand_str }
    },
    {
        MVM_OP_param_rn2_o,
        "param_rn2_o",
        3,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_str }
    },
    {
        MVM_OP_param_on2_i,
        "param_on2_i",
        4,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_str, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_param_on2_n,
        "param_on2_n",
        4,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_str, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_param_on2_s,
        "param_on2_s",
        4,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_str, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_param_on2_o,
        "param_on2_o",
        4,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_osrpoint,
        "osrpoint",
        0,
        0,
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_nativecallcast,
        "nativecallcast",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_spawnprocasync,
        "spawnprocasync",
        7,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_killprocasync,
        "killprocasync",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_startprofile,
        "startprofile",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_endprofile,
        "endprofile",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_objectid,
        "objectid",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_settypefinalize,
        "settypefinalize",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_force_gc,
        "force_gc",
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_nativecallglobal,
        "nativecallglobal",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_32,
        "DEPRECATED_32",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_setparameterizer,
        "setparameterizer",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_parameterizetype,
        "parameterizetype",
        3,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_typeparameterized,
        "typeparameterized",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_typeparameters,
        "typeparameters",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_typeparameterat,
        "typeparameterat",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_readlink,
        "readlink",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_lstat,
        "lstat",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_iscont_i,
        "iscont_i",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_iscont_n,
        "iscont_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_iscont_s,
        "iscont_s",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_assign_i,
        "assign_i",
        2,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_assign_n,
        "assign_n",
        2,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_assign_s,
        "assign_s",
        2,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_decont_i,
        "decont_i",
        2,
        1,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decont_n,
        "decont_n",
        2,
        1,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decont_s,
        "decont_s",
        2,
        1,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getrusage,
        "getrusage",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_threadlockcount,
        "threadlockcount",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bindcomplete,
        "bindcomplete",
        0,
        0,
        2,
        1,
        0,
        0,
        1,
        1,
        0,
        0,
        1,
        { 0 }
    },
    {
        MVM_OP_getlexref_i,
        "getlexref_i",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_int64 }
    },
    {
        MVM_OP_getlexref_n,
        "getlexref_n",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_num64 }
    },
    {
        MVM_OP_getlexref_s,
        "getlexref_s",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_str }
    },
    {
        MVM_OP_getlexref_ni,
        "getlexref_ni",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str }
    },
    {
        MVM_OP_getlexref_nn,
        "getlexref_nn",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str }
    },
    {
        MVM_OP_getlexref_ns,
        "getlexref_ns",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str }
    },
    {
        MVM_OP_atposref_i,
        "atposref_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atposref_n,
        "atposref_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atposref_s,
        "atposref_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_getattrref_i,
        "getattrref_i",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_getattrref_n,
        "getattrref_n",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_getattrref_s,
        "getattrref_s",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_getattrsref_i,
        "getattrsref_i",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getattrsref_n,
        "getattrsref_n",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getattrsref_s,
        "getattrsref_s",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_nativecallsizeof,
        "nativecallsizeof",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_encodenorm,
        "encodenorm",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_normalizecodes,
        "normalizecodes",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_strfromcodes,
        "strfromcodes",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_strtocodes,
        "strtocodes",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_getcodelocation,
        "getcodelocation",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_eqatim_s,
        "eqatim_s",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_ordbaseat,
        "ordbaseat",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_neverrepossess,
        "neverrepossess",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_scdisclaim,
        "scdisclaim",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_setthreadname,
        "setthreadname",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_atpos2d_i,
        "atpos2d_i",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atpos2d_n,
        "atpos2d_n",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atpos2d_s,
        "atpos2d_s",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atpos2d_o,
        "atpos2d_o",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atpos3d_i,
        "atpos3d_i",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atpos3d_n,
        "atpos3d_n",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atpos3d_s,
        "atpos3d_s",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atpos3d_o,
        "atpos3d_o",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atposnd_i,
        "atposnd_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_atposnd_n,
        "atposnd_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_atposnd_s,
        "atposnd_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_atposnd_o,
        "atposnd_o",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bindpos2d_i,
        "bindpos2d_i",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bindpos2d_n,
        "bindpos2d_n",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_bindpos2d_s,
        "bindpos2d_s",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindpos2d_o,
        "bindpos2d_o",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bindpos3d_i,
        "bindpos3d_i",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bindpos3d_n,
        "bindpos3d_n",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_bindpos3d_s,
        "bindpos3d_s",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindpos3d_o,
        "bindpos3d_o",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bindposnd_i,
        "bindposnd_i",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_bindposnd_n,
        "bindposnd_n",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_bindposnd_s,
        "bindposnd_s",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_bindposnd_o,
        "bindposnd_o",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_dimensions,
        "dimensions",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_setdimensions,
        "setdimensions",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_numdimensions,
        "numdimensions",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_ctxcode,
        "ctxcode",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_isrwcont,
        "isrwcont",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_fc,
        "fc",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_DEPRECATED_25,
        "DEPRECATED_25",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_atpos3d_u,
        "atpos3d_u",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_encoderep,
        "encoderep",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_istty_fh,
        "istty_fh",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_multidimref_i,
        "multidimref_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_multidimref_n,
        "multidimref_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_multidimref_s,
        "multidimref_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_fileno_fh,
        "fileno_fh",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_asyncudp,
        "asyncudp",
        7,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bindpos2d_u,
        "bindpos2d_u",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_asyncwritebytesto,
        "asyncwritebytesto",
        8,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_objprimbits,
        "objprimbits",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_objprimunsigned,
        "objprimunsigned",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_ctxnt,
        "ctxnt",
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_chown_f,
        "chown_f",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_bindattr_u,
        "bindattr_u",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_int16 }
    },
    {
        MVM_OP_bindattrs_u,
        "bindattrs_u",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_getlexref_i32,
        "getlexref_i32",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_int32 }
    },
    {
        MVM_OP_getlexref_i16,
        "getlexref_i16",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_int16 }
    },
    {
        MVM_OP_getlexref_i8,
        "getlexref_i8",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_int8 }
    },
    {
        MVM_OP_getlexref_n32,
        "getlexref_n32",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_num32 }
    },
    {
        MVM_OP_box_u,
        "box_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_unbox_u,
        "unbox_u",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_coerce_iu,
        "coerce_iu",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_coerce_ui,
        "coerce_ui",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_coerce_nu,
        "coerce_nu",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_coerce_un,
        "coerce_un",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_decont_u,
        "decont_u",
        2,
        1,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bnot_u,
        "bnot_u",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_getattr_u,
        "getattr_u",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_getattrs_u,
        "getattrs_u",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_getattrref_u,
        "getattrref_u",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_getlexref_u,
        "getlexref_u",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_uint64 }
    },
    {
        MVM_OP_getlexref_u32,
        "getlexref_u32",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_uint32 }
    },
    {
        MVM_OP_getlexref_u16,
        "getlexref_u16",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_uint16 }
    },
    {
        MVM_OP_getlexref_u8,
        "getlexref_u8",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_uint8 }
    },
    {
        MVM_OP_param_rp_u,
        "param_rp_u",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_int16 }
    },
    {
        MVM_OP_param_op_u,
        "param_op_u",
        3,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_int16, MVM_operand_ins }
    },
    {
        MVM_OP_param_rn_u,
        "param_rn_u",
        2,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_str }
    },
    {
        MVM_OP_param_on_u,
        "param_on_u",
        3,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_param_rn2_u,
        "param_rn2_u",
        3,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_str, MVM_operand_str }
    },
    {
        MVM_OP_param_on2_u,
        "param_on2_u",
        4,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_str, MVM_operand_str, MVM_operand_ins }
    },
    {
        MVM_OP_stat_time,
        "stat_time",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_lstat_time,
        "lstat_time",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_setdebugtypename,
        "setdebugtypename",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_loadbytecodebuffer,
        "loadbytecodebuffer",
        1,
        0,
        3,
        1,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_loadbytecodefh,
        "loadbytecodefh",
        2,
        0,
        3,
        1,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_throwpayloadlex,
        "throwpayloadlex",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_throwpayloadlexcaller,
        "throwpayloadlexcaller",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_lastexpayload,
        "lastexpayload",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_cancelnotify,
        "cancelnotify",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decoderconfigure,
        "decoderconfigure",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decodersetlineseps,
        "decodersetlineseps",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decoderaddbytes,
        "decoderaddbytes",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decodertakechars,
        "decodertakechars",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_decodertakeallchars,
        "decodertakeallchars",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decodertakeavailablechars,
        "decodertakeavailablechars",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decodertakeline,
        "decodertakeline",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_decoderbytesavailable,
        "decoderbytesavailable",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_decodertakebytes,
        "decodertakebytes",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_decoderempty,
        "decoderempty",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_indexingoptimized,
        "indexingoptimized",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_captureinnerlex,
        "captureinnerlex",
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_unicmp_s,
        "unicmp_s",
        6,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_DEPRECATED_92,
        "DEPRECATED_92",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_strfromname,
        "strfromname",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_indexic_s,
        "indexic_s",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_getport_sk,
        "getport_sk",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_cpucores,
        "cpucores",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_eqaticim_s,
        "eqaticim_s",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_indexicim_s,
        "indexicim_s",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_decodertakecharseof,
        "decodertakecharseof",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_indexim_s,
        "indexim_s",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_cas_o,
        "cas_o",
        4,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_cas_i,
        "cas_i",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atomicinc_i,
        "atomicinc_i",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_atomicdec_i,
        "atomicdec_i",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_atomicadd_i,
        "atomicadd_i",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atomicload_o,
        "atomicload_o",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_atomicload_i,
        "atomicload_i",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_atomicstore_o,
        "atomicstore_o",
        2,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_atomicstore_i,
        "atomicstore_i",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_barrierfull,
        "barrierfull",
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_coveragecontrol,
        "coveragecontrol",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_DEPRECATED_111,
        "DEPRECATED_111",
        2,
        0,
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_112,
        "DEPRECATED_112",
        3,
        0,
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_113,
        "DEPRECATED_113",
        3,
        0,
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_114,
        "DEPRECATED_114",
        3,
        0,
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_115,
        "DEPRECATED_115",
        3,
        0,
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_107,
        "DEPRECATED_107",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int16 }
    },
    {
        MVM_OP_DEPRECATED_108,
        "DEPRECATED_108",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_int16 }
    },
    {
        MVM_OP_DEPRECATED_109,
        "DEPRECATED_109",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int16 }
    },
    {
        MVM_OP_DEPRECATED_110,
        "DEPRECATED_110",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int16 }
    },
    {
        MVM_OP_coerce_II,
        "coerce_II",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_encoderepconf,
        "encoderepconf",
        6,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_encodeconf,
        "encodeconf",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_decodeconf,
        "decodeconf",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_decoderepconf,
        "decoderepconf",
        5,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_getppid,
        "getppid",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_getsignals,
        "getsignals",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_slice,
        "slice",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_DEPRECATED_38,
        "DEPRECATED_38",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_39,
        "DEPRECATED_39",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str }
    },
    {
        MVM_OP_DEPRECATED_81,
        "DEPRECATED_81",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_41,
        "DEPRECATED_41",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_42,
        "DEPRECATED_42",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_43,
        "DEPRECATED_43",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_44,
        "DEPRECATED_44",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_str }
    },
    {
        MVM_OP_atomicbindattr_o,
        "atomicbindattr_o",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_casattr_o,
        "casattr_o",
        6,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_atkey_u,
        "atkey_u",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_coerce_us,
        "coerce_us",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_DEPRECATED_45,
        "DEPRECATED_45",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_hllbool,
        "hllbool",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_hllboolfor,
        "hllboolfor",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_decodelocaltime,
        "decodelocaltime",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_fork,
        "fork",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_buffertocu,
        "buffertocu",
        2,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_writeint,
        "writeint",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_writeuint,
        "writeuint",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_writenum,
        "writenum",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_serializetobuf,
        "serializetobuf",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_readint,
        "readint",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_readuint,
        "readuint",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_readnum,
        "readnum",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_DEPRECATED_68,
        "DEPRECATED_68",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_uname,
        "uname",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_freemem,
        "freemem",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_totalmem,
        "totalmem",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_DEPRECATED_93,
        "DEPRECATED_93",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_DEPRECATED_94,
        "DEPRECATED_94",
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_time,
        "time",
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_dispatch_v,
        "dispatch_v",
        2,
        0,
        10,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        { MVM_operand_str, MVM_operand_callsite }
    },
    {
        MVM_OP_dispatch_i,
        "dispatch_i",
        3,
        0,
        10,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_str, MVM_operand_callsite }
    },
    {
        MVM_OP_dispatch_n,
        "dispatch_n",
        3,
        0,
        10,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_str, MVM_operand_callsite }
    },
    {
        MVM_OP_dispatch_s,
        "dispatch_s",
        3,
        0,
        10,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_str, MVM_operand_callsite }
    },
    {
        MVM_OP_dispatch_o,
        "dispatch_o",
        3,
        0,
        11,
        1,
        0,
        1,
        0,
        0,
        0,
        1,
        1,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_callsite }
    },
    {
        MVM_OP_gettypehllrole,
        "gettypehllrole",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_bindpos_u,
        "bindpos_u",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_atpos_u,
        "atpos_u",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_atposref_u,
        "atposref_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_iscont_u,
        "iscont_u",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_assign_u,
        "assign_u",
        2,
        0,
        2,
        1,
        0,
        0,
        0,
        1,
        0,
        1,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_uint64 }
    },
    {
        MVM_OP_captureposarg_u,
        "captureposarg_u",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_sp_guard,
        "sp_guard",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_guardconc,
        "sp_guardconc",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_guardtype,
        "sp_guardtype",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_guardhll,
        "sp_guardhll",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_guardsf,
        "sp_guardsf",
        3,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_guardsfouter,
        "sp_guardsfouter",
        3,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_guardobj,
        "sp_guardobj",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_guardnotobj,
        "sp_guardnotobj",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_guardjustconc,
        "sp_guardjustconc",
        3,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_guardjusttype,
        "sp_guardjusttype",
        3,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_guardnonzero,
        "sp_guardnonzero",
        3,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_rebless,
        "sp_rebless",
        4,
        0,
        0,
        1,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_decont,
        "sp_decont",
        2,
        1,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_getlex_o,
        "sp_getlex_o",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_lex | MVM_operand_obj }
    },
    {
        MVM_OP_sp_getlex_ins,
        "sp_getlex_ins",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_type_var, MVM_operand_read_lex | MVM_operand_type_var }
    },
    {
        MVM_OP_sp_getlex_no,
        "sp_getlex_no",
        2,
        1,
        0,
        0,
        0,
        0,
        1,
        3,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str }
    },
    {
        MVM_OP_sp_bindlex_in,
        "sp_bindlex_in",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_lex | MVM_operand_type_var, MVM_operand_read_reg | MVM_operand_type_var }
    },
    {
        MVM_OP_sp_bindlex_os,
        "sp_bindlex_os",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_lex | MVM_operand_type_var, MVM_operand_read_reg | MVM_operand_type_var }
    },
    {
        MVM_OP_sp_getlexstatic_o,
        "sp_getlexstatic_o",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_str, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_assertparamcheck,
        "sp_assertparamcheck",
        3,
        0,
        0,
        1,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_int64, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_bindcomplete,
        "sp_bindcomplete",
        2,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_istype,
        "sp_istype",
        5,
        1,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_dispatch_v,
        "sp_dispatch_v",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_str, MVM_operand_callsite, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_dispatch_i,
        "sp_dispatch_i",
        5,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_str, MVM_operand_callsite, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_dispatch_u,
        "sp_dispatch_u",
        5,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_str, MVM_operand_callsite, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_dispatch_n,
        "sp_dispatch_n",
        5,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_str, MVM_operand_callsite, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_dispatch_s,
        "sp_dispatch_s",
        5,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_str, MVM_operand_callsite, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_dispatch_o,
        "sp_dispatch_o",
        5,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_str, MVM_operand_callsite, MVM_operand_spesh_slot, MVM_operand_uint32 }
    },
    {
        MVM_OP_sp_getarg_o,
        "sp_getarg_o",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_getarg_i,
        "sp_getarg_i",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_getarg_n,
        "sp_getarg_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_getarg_s,
        "sp_getarg_s",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_paramnamesused,
        "sp_paramnamesused",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_spesh_slot }
    },
    {
        MVM_OP_sp_getspeshslot,
        "sp_getspeshslot",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_spesh_slot }
    },
    {
        MVM_OP_sp_fastcreate,
        "sp_fastcreate",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot }
    },
    {
        MVM_OP_sp_get_o,
        "sp_get_o",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_get_i64,
        "sp_get_i64",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_get_i32,
        "sp_get_i32",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_get_i16,
        "sp_get_i16",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_get_i8,
        "sp_get_i8",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_get_n,
        "sp_get_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_get_s,
        "sp_get_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_bind_o,
        "sp_bind_o",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_bind_i64,
        "sp_bind_i64",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_sp_bind_i32,
        "sp_bind_i32",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_sp_bind_i16,
        "sp_bind_i16",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_sp_bind_i8,
        "sp_bind_i8",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_sp_bind_n,
        "sp_bind_n",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_sp_bind_s,
        "sp_bind_s",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_sp_bind_s_nowb,
        "sp_bind_s_nowb",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_sp_p6oget_o,
        "sp_p6oget_o",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_p6ogetvt_o,
        "sp_p6ogetvt_o",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot }
    },
    {
        MVM_OP_sp_p6ogetvc_o,
        "sp_p6ogetvc_o",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot }
    },
    {
        MVM_OP_sp_p6oget_i,
        "sp_p6oget_i",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_p6oget_n,
        "sp_p6oget_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_p6oget_s,
        "sp_p6oget_s",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_p6oget_bi,
        "sp_p6oget_bi",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_p6obind_o,
        "sp_p6obind_o",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_p6obind_i,
        "sp_p6obind_i",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_sp_p6obind_n,
        "sp_p6obind_n",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_sp_p6obind_s,
        "sp_p6obind_s",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_str }
    },
    {
        MVM_OP_sp_p6oget_i32,
        "sp_p6oget_i32",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_p6obind_i32,
        "sp_p6obind_i32",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_sp_getvt_o,
        "sp_getvt_o",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot }
    },
    {
        MVM_OP_sp_getvc_o,
        "sp_getvc_o",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot }
    },
    {
        MVM_OP_sp_fastbox_i,
        "sp_fastbox_i",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_sp_fastbox_bi,
        "sp_fastbox_bi",
        5,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64 }
    },
    {
        MVM_OP_sp_fastbox_i_ic,
        "sp_fastbox_i_ic",
        6,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_fastbox_bi_ic,
        "sp_fastbox_bi_ic",
        6,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_deref_get_i64,
        "sp_deref_get_i64",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_deref_get_n,
        "sp_deref_get_n",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_deref_bind_i64,
        "sp_deref_bind_i64",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_int64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_deref_bind_n,
        "sp_deref_bind_n",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_num64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_getlexvia_o,
        "sp_getlexvia_o",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_getlexvia_ins,
        "sp_getlexvia_ins",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_type_var, MVM_operand_int16, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_bindlexvia_os,
        "sp_bindlexvia_os",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_int16, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_bindlexvia_in,
        "sp_bindlexvia_in",
        4,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_int16, MVM_operand_int16, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_getstringfrom,
        "sp_getstringfrom",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_spesh_slot, MVM_operand_int32 }
    },
    {
        MVM_OP_sp_getwvalfrom,
        "sp_getwvalfrom",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_spesh_slot, MVM_operand_int64 }
    },
    {
        MVM_OP_sp_jit_enter,
        "sp_jit_enter",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_istrue_n,
        "sp_istrue_n",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_num64 }
    },
    {
        MVM_OP_sp_boolify_iter,
        "sp_boolify_iter",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_boolify_iter_arr,
        "sp_boolify_iter_arr",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_boolify_iter_hash,
        "sp_boolify_iter_hash",
        2,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_cas_o,
        "sp_cas_o",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_atomicload_o,
        "sp_atomicload_o",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_atomicstore_o,
        "sp_atomicstore_o",
        2,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_sp_add_I,
        "sp_add_I",
        7,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_sub_I,
        "sp_sub_I",
        7,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_mul_I,
        "sp_mul_I",
        7,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_spesh_slot, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_bool_I,
        "sp_bool_I",
        3,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_runbytecode_v,
        "sp_runbytecode_v",
        3,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_runbytecode_i,
        "sp_runbytecode_i",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_runbytecode_u,
        "sp_runbytecode_u",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_runbytecode_n,
        "sp_runbytecode_n",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_runbytecode_s,
        "sp_runbytecode_s",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_runbytecode_o,
        "sp_runbytecode_o",
        4,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64, MVM_operand_int16 }
    },
    {
        MVM_OP_sp_runcfunc_v,
        "sp_runcfunc_v",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_runcfunc_i,
        "sp_runcfunc_i",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_runcfunc_u,
        "sp_runcfunc_u",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_runcfunc_n,
        "sp_runcfunc_n",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_runcfunc_s,
        "sp_runcfunc_s",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_runcfunc_o,
        "sp_runcfunc_o",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_runnativecall_v,
        "sp_runnativecall_v",
        2,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_runnativecall_i,
        "sp_runnativecall_i",
        3,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_int64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_runnativecall_u,
        "sp_runnativecall_u",
        3,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_uint64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_runnativecall_n,
        "sp_runnativecall_n",
        3,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_num64, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_runnativecall_s,
        "sp_runnativecall_s",
        3,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_str, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_runnativecall_o,
        "sp_runnativecall_o",
        3,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_read_reg | MVM_operand_obj, MVM_operand_uint64 }
    },
    {
        MVM_OP_sp_resumption,
        "sp_resumption",
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_write_reg | MVM_operand_obj, MVM_operand_uint16, MVM_operand_uint16 }
    },
    {
        MVM_OP_prof_enter,
        "prof_enter",
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_prof_enterspesh,
        "prof_enterspesh",
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_prof_enterinline,
        "prof_enterinline",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_spesh_slot }
    },
    {
        MVM_OP_prof_enternative,
        "prof_enternative",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_prof_exit,
        "prof_exit",
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { 0 }
    },
    {
        MVM_OP_prof_allocated,
        "prof_allocated",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj }
    },
    {
        MVM_OP_prof_replaced,
        "prof_replaced",
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_spesh_slot }
    },
    {
        MVM_OP_ctw_check,
        "ctw_check",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_read_reg | MVM_operand_obj, MVM_operand_int16 }
    },
    {
        MVM_OP_coverage_log,
        "coverage_log",
        4,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        { MVM_operand_str, MVM_operand_int32, MVM_operand_int32, MVM_operand_int64 }
    },
    {
        MVM_OP_breakpoint,
        "breakpoint",
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        0,
        0,
        0,
        { MVM_operand_int32, MVM_operand_int32 }
    },
};

static const unsigned short MVM_op_counts = 958;

static const MVMuint16 last_op_allowed = 837;

static const MVMuint8 MVM_op_allowed_in_confprog[] = {
    0xD1, 0x1, 0x80, 0x3,
    0x3, 0x0, 0x0, 0xFF,
    0xFF, 0xF0, 0xCB, 0xFF,
    0x6F, 0xC0, 0xA8, 0x7F,
    0x0, 0x0, 0x0, 0x0,
    0x0, 0x0, 0x0, 0x0,
    0xF0, 0xFF, 0xEF, 0x7,
    0xF0, 0x1, 0x0, 0x1,
    0xD, 0x5, 0x40, 0x80,
    0x16, 0x0, 0x0, 0x10,
    0xA0, 0x0, 0x0, 0xF8,
    0x0, 0x0, 0x0, 0x0,
    0x40, 0x0, 0x0, 0x0,
    0x0, 0x0, 0x0, 0x0,
    0x0, 0x0, 0x0, 0x0,
    0x0, 0x82, 0x80, 0x41,
    0x7, 0x1, 0x20, 0x0,
    0x0, 0x0, 0x0, 0x0,
    0x0, 0x0, 0x60, 0x0,
    0x0, 0x0, 0x1C, 0x0,
    0x0, 0x0, 0x4, 0x0,
    0x0, 0x0, 0x0, 0x0,
    0x0, 0x0, 0x4, 0x0,
    0x0, 0x0, 0x0, 0x0,
    0x0, 0x0, 0x0, 0x0,
    0x0, 0x0, 0x0, 0x2,
    0xA,};

MVM_PUBLIC const MVMOpInfo * MVM_op_get_op(unsigned short op) {
    if (op >= MVM_op_counts)
        return NULL;
    return &MVM_op_infos[op];
}

MVM_PUBLIC MVMuint8 MVM_op_is_allowed_in_confprog(unsigned short op) {
    if (op > last_op_allowed)
        return 0;
    return !!(MVM_op_allowed_in_confprog[op / 8] & (1 << (op % 8)));
}

MVM_PUBLIC const char *MVM_op_get_mark(unsigned short op) {
    if (op > 838 && op < MVM_OP_EXT_BASE) {
        return ".s";
    } else if (op == 23) {
        return ".j";
    } else if (op == 34) {
        return ":j";
    } else if (op >= 51 && op < 56) {
        return ".r";
    } else if (op == 127) {
        return "+a";
    } else if (op >= 128 && op < 135) {
        return "*a";
    } else if (op >= 135 && op < 140) {
        return "-a";
    } else if (op >= 141 && op < 149) {
        return ".p";
    } else if (op == 157) {
        return ".p";
    } else if (op == 473) {
        return ".d";
    } else if (op == 496) {
        return ".r";
    } else if (op >= 778 && op < 783) {
        return "-a";
    } else if (op == 796) {
        return "-a";
    } else if (op >= 826 && op < 831) {
        return ".d";
    } else if (op >= MVM_OP_EXT_BASE) {
        return ".x";
    }
    return "  ";
}


